/*
 *  gifford W. scott
 *  Copyright (C) 2005 LVL7 Systems, Inc.  All rights reserved.
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <asm/io.h>
#include <asm/bootinfo.h>

extern u8 bcm47xx_et_macaddr[2][18];


char arcs_cmdline[COMMAND_LINE_SIZE]; /* poorly named global ... */
static char txtSpace[] = " ";


/*
  DESCR:  copy text and additional space char (' ') to buffer
  PARMS:  pBfr - ptr to buffer
          pBfrTerm - ptr to terminating element of buffer
          pText - ptr to text to append to buffer
  RETURN: ptr to terminating NIL char in buffer
  PRE:    - all ptrs valid (non-NULL)
 */
static char* cpyTxtWithSpace(char* pBfr, char* const pBfrTerm, char* pTxt)
{
  /* calc len of arg (does not include terminating '\0') */
  size_t argLen = strlen(pTxt);

  /* if enough room in buffer for arg and appended space char: */
  if ((argLen + sizeof(txtSpace)) <= (pBfrTerm - pBfr))
  {
    /* copy arg & appended space char to buffer */
    memcpy( pBfr, pTxt, argLen );
    pBfr += argLen;
    memcpy( pBfr, txtSpace, sizeof(txtSpace) );

    /* point to NIL char */
    pBfr += sizeof(txtSpace) - 1;
  }

  return pBfr;
}


void prom_free_prom_memory(void)
{
}


/*
  DESCR:  'massage' argument command line presented on boot-up
  PARMS:  argc - boot argument count
          argv - argument vector; ptr to array of [0 ... (argc-1)] ptrs
                 to boot arguments
          envp - environment vector; ptr to array of environment variables
  RETURN: zero
 */
int __init prom_init(int argc, char** argv, char **envp)
{
	char* pBfr = arcs_cmdline;
  char* const pBfrTerm = arcs_cmdline + COMMAND_LINE_SIZE;
  char** const argvTerm = argv + argc;
  unsigned long val;

  /*
   * scan each command line argument & place in buffer w/appended space
   * - while (arg ptrs remain to be processed)
   *         AND (room in text buffer)
   */
  while (
         (argv < argvTerm)
         && (pBfr < pBfrTerm)
        )
  {
    /*
     * adjust loop vars:
     * - if arg ptr valid, add arg text & appended space to buffer
     * - inc argv to point at next arg ptr
     */
    if (*argv != NULL)
    {
      pBfr = cpyTxtWithSpace(pBfr, pBfrTerm, *argv);
    }
    argv += 1;
	}

  /* scan each env var */
  while (*envp != NULL)
  {
    if ((pBfr = strstr(*envp, "memsize=")) != NULL)
    {
      sscanf(pBfr, "memsize=%ld", &val);
      add_memory_region(0, val << 20 , BOOT_MEM_RAM);
    }

    else if ((pBfr = strstr(*envp, "baudrate=")) != NULL)
    {
      sscanf(pBfr, "baudrate=%ld", &val);
      if ((pBfr = strstr(arcs_cmdline, "console=")) == NULL)
      {
        /*JWL -- need better string handling... */
        sprintf(arcs_cmdline, "%s console=ttyS0,%ld", arcs_cmdline, val);
      }
		}

    else if ((pBfr = strstr(*envp, "ethaddr=")) != NULL)
    {
      memcpy(bcm47xx_et_macaddr[0], pBfr+strlen("ethaddr="),
             sizeof(bcm47xx_et_macaddr[0])-1);
      bcm47xx_et_macaddr[0][sizeof(bcm47xx_et_macaddr[0])-1] = '\0';
    }

#ifdef CONFIG_BLK_DEV_INITRD
    /* vars to store the limits of the initial ramdisk */
    else if ((pBfr = strstr(*envp, "initrd_start=")) != NULL)
    {
      sscanf(pBfr, "initrd_start=0x%lx", &val);
      initrd_start = KSEG0ADDR(val);
    }

    else if ((pBfr = strstr(*envp, "initrd_size=")) != NULL)
    {
      /* conversion doesn't work if first letter is not a decimal digit... (?) */
      *(pBfr+strlen("initrd_size=0")) = '0';
      sscanf(pBfr, "initrd_size=%lx", &val);
      initrd_end = KSEG0ADDR(initrd_start + val);
    }
#endif /* CONFIG_BLK_DEV_INITRD */

    /* adjust loop var */
    envp += 1;
  }

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start != 0)
  {
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	}
#endif /* CONFIG_BLK_DEV_INITRD */

	return 0;
}
