/*
 *  John W. Linville
 *  Copyright (C) 2005 LVL7 Systems, Inc.  All rights reserved.
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/serial.h>
#include <linux/types.h>
#include <linux/string.h>       /* for memset */

#include <asm/reboot.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/pgtable.h>
#include <asm/processor.h>
#include <asm/reboot.h>
#include <asm/system.h>

extern u32 bcm47xx_idmask;

u8  bcm47xx_et_macaddr[2][18];
u32 bcm47xx_et_phyaddr[2];
u32 bcm47xx_et_mdcport[2];

extern struct resource ioport_resource;

void fsm73xxs_restart(char *command)
{
#if 0 /*JWL*/
	/* board reset -- JWL */
	*((u32 *)0xb8000064) = 0x00000001;
	while (1);
#else
	/* Disable BCM47xx caches */
	*((u32 *)0xff400000) = 0; /* Disable PFC (aka RAC) */
	clear_cp0_diag(0xc0000000); /* Disable I-cache & D-cache */
	
	set_cp0_status(ST0_BEV | ST0_ERL);
	change_cp0_config(CONF_CM_CMASK, CONF_CM_UNCACHED);
	flush_cache_all();
	write_32bit_cp0_register(CP0_WIRED, 0);
	__asm__ __volatile__("jr\t%0"::"r"(0xbfc00000));
#endif
}

void fsm73xxs_halt(void)
{
	printk(KERN_NOTICE "\n** You may safely turn-off the system. **\n");
	while (1);
}

void fsm73xxs_power_off(void)
{
	fsm73xxs_halt();
}

static void __init fsm73xxs_time_init(void)
{
	/* 100 MHz (SB clock) */
	mips_counter_frequency = 100000000;
}

extern int setup_irq(unsigned int irq, struct irqaction *irqaction);
static void __init fsm73xxs_timer_setup(struct irqaction *irq)
{
	unsigned int count;

	setup_irq(7, irq);

	/* to generate the first CPU timer interrupt */
	count = read_32bit_cp0_register(CP0_COUNT);
	write_32bit_cp0_register(CP0_COMPARE, count + 8000);
}

void fsm73xxs_setup(void)
{
	printk("Netgear FSM73xxS setup...\n");

	if ((*((u32 *)0xb8000060) & 0x00000020) == 0) {

		bcm47xx_idmask = 0x00004001;

	} else {

		bcm47xx_idmask = 0x00006001;

	}

	bcm47xx_et_phyaddr[0] = 2;
	bcm47xx_et_mdcport[0] = 0;

	board_time_init = fsm73xxs_time_init;
	board_timer_setup = fsm73xxs_timer_setup;

	_machine_restart = fsm73xxs_restart;
	_machine_halt = fsm73xxs_halt;
	_machine_power_off = fsm73xxs_power_off;

	set_io_port_base(0);
	ioport_resource.start = 0xb8000000;
	ioport_resource.end = 0xbfffffff;
}
