/*
 *  John W. Linville, linville@lvl7.com
 *  Copyright (C) 2003 LVL7 Systems, Inc.  All rights reserved.
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/kernel.h>

#include <asm/io.h>
#include <asm/bootinfo.h>

#ifdef CONFIG_BLK_DEV_INITRD
extern unsigned long initrd_start, initrd_end;
#endif

char arcs_cmdline[COMMAND_LINE_SIZE] = { 0, }; /* poorly named, but... */

extern u8 bcm47xx_et_macaddr[2][18];

void prom_free_prom_memory(void) {}

int __init prom_init(int argc, char **argv, char **envp)
{
	unsigned long val;
	int i;
	char *argp;

	argp = arcs_cmdline;

	for (i=0; i<argc; i++) {

		if (argv[i] != NULL) {

			if (argp != arcs_cmdline) {

				*argp++ = ' ';
				*argp = '\0';

			}

			strcat(argp, argv[i]);
			argp += strlen(argv[i]);

		}

	}

	if (envp[i=0] != NULL) {

		do {

			if ((argp = strstr(envp[i], "memsize=")) != NULL) {

				sscanf(argp, "memsize=%d", &val);
				add_memory_region(0, val << 20 , BOOT_MEM_RAM);
				continue;

			}

			if ((argp = strstr(envp[i], "baudrate=")) != NULL) {

				sscanf(argp, "baudrate=%d", &val);
				if ((argp = strstr(arcs_cmdline, "console="))
				 == NULL) {

/*JWL -- need better string handling... */
					sprintf(arcs_cmdline,
					        "%s console=ttyS0,%d",
					        arcs_cmdline, val);

				}
				continue;

			}

			if ((argp = strstr(envp[i], "ethaddr=")) != NULL) {

				memcpy(bcm47xx_et_macaddr[0],
				       argp+strlen("ethaddr="),
				       sizeof(bcm47xx_et_macaddr[0])-1);
				bcm47xx_et_macaddr[0][sizeof(bcm47xx_et_macaddr[0])-1] = '\0';
				continue;

			}

#ifdef CONFIG_BLK_DEV_INITRD
			if ((argp = strstr(envp[i], "initrd_start=")) != NULL) {

				sscanf(argp, "initrd_start=0x%x", &val);
				initrd_start = KSEG0ADDR(val);
				continue;

			}

			if ((argp = strstr(envp[i], "initrd_size=")) != NULL) {

				/* conversion doesn't work if first
				   letter is not a decimal digit... (?) */
				*(argp+strlen("initrd_size=0")) = '0';
				sscanf(argp, "initrd_size=%x", &val);
				initrd_end = KSEG0ADDR(initrd_start + val);
				continue;

			}
#endif

		} while (envp[++i] != NULL);
	}

	if (initrd_start != 0) {

		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);

	}

	return 0;
}
