/*
 * $Id: fsm73xx.c,v 1.1 2003/09/23 21:30:12 linville Exp $
 *
 * Normal mappings of chips in physical memory
 *
 * Modified for Netgear FSM73xx partitions... -- JWL
 */

/* partition support */
#define HAVE_PARTITIONS

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>
#ifdef HAVE_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define WINDOW_ADDR	0xff000000
#define WINDOW_SIZE	0x01000000
#define BUSWIDTH	1

#define APP_START	0x00000000
#define APP_LEN		0x00de0000

#define SYSTEM_START	0x00de0000
#define SYSTEM_LEN	0x00120000

#define U_BOOT_START	0x00f00000
#define U_BOOT_LEN	0x00020000

#define ENV_START	0x00f20000
#define ENV_LEN		0x00020000

#define VPD_START	0x00f40000
#define VPD_LEN		0x00020000

#define SYSTEM_2_START	0x00f60000
#define SYSTEM_2_LEN	0x000a0000

#ifdef HAVE_PARTITIONS
static struct mtd_partition fsm73xx_partitions[] =
{
	/* app filesystem */
	{
		      name: "Application Filesystem",
		    offset: APP_START,
		      size: APP_LEN,
		mask_flags: 0
	},
	/* system (kernel + ramdisk) */
	{
		      name: "System (Kernel + Ramdisk)",
		    offset: SYSTEM_START,
		      size: SYSTEM_LEN,
		mask_flags: 0
	},
	/* u-boot */
	{
		      name: "U-Boot",
		    offset: U_BOOT_START,
		      size: U_BOOT_LEN,
		mask_flags: 0
	},
	/* u-boot environment */
	{
		      name: "U-Boot Environment",
		    offset: ENV_START,
		      size: ENV_LEN,
		mask_flags: 0
	},
	/* VPD */
	{
		      name: "FASTPATH Vital Product Data",
		    offset: VPD_START,
		      size: VPD_LEN,
		mask_flags: 0
	},
	/* system (part 2) */
	{
		      name: "System (part 2)",
		    offset: SYSTEM_2_START,
		      size: SYSTEM_2_LEN,
		mask_flags: 0
	}
};

static struct mtd_partition fsm73xx_partitions_2[] =
{
	/* app filesystem */
	{
		      name: "Application Filesystem",
		    offset: APP_START,
		      size: APP_LEN - (WINDOW_SIZE / 2),
		mask_flags: 0
	},
	/* system (kernel + ramdisk) */
	{
		      name: "System (Kernel + Ramdisk)",
		    offset: SYSTEM_START - (WINDOW_SIZE / 2),
		      size: SYSTEM_LEN,
		mask_flags: 0
	},
	/* u-boot */
	{
		      name: "U-Boot",
		    offset: U_BOOT_START - (WINDOW_SIZE / 2),
		      size: U_BOOT_LEN,
		mask_flags: 0
	},
	/* u-boot environment */
	{
		      name: "U-Boot Environment",
		    offset: ENV_START - (WINDOW_SIZE / 2),
		      size: ENV_LEN,
		mask_flags: 0
	},
	/* VPD */
	{
		      name: "FASTPATH Vital Product Data",
		    offset: VPD_START - (WINDOW_SIZE / 2),
		      size: VPD_LEN,
		mask_flags: 0
	},
	/* system (part 2) */
	{
		      name: "System (part 2)",
		    offset: SYSTEM_2_START - (WINDOW_SIZE / 2),
		      size: SYSTEM_2_LEN,
		mask_flags: 0
	}
};

#define NUM_PARTITIONS	(sizeof(fsm73xx_partitions) \
                       / sizeof(struct mtd_partition))
#endif

static struct mtd_info *mymtd;

__u8 fsm73xx_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 fsm73xx_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 fsm73xx_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void fsm73xx_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void fsm73xx_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void fsm73xx_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void fsm73xx_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void fsm73xx_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info fsm73xx_map = {
	name: "Physically mapped flash",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: fsm73xx_read8,
	read16: fsm73xx_read16,
	read32: fsm73xx_read32,
	copy_from: fsm73xx_copy_from,
	write8: fsm73xx_write8,
	write16: fsm73xx_write16,
	write32: fsm73xx_write32,
	copy_to: fsm73xx_copy_to
};

int __init init_fsm73xx(void)
{
       	printk(KERN_NOTICE "fsm73xx flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
	fsm73xx_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);

	if (!fsm73xx_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	mymtd = do_map_probe("cfi_probe", &fsm73xx_map);
	if (mymtd) {
		mymtd->module = THIS_MODULE;

#if defined(HAVE_PARTITIONS)
		add_mtd_partitions(mymtd, fsm73xx_partitions,
		                   NUM_PARTITIONS);
#else
		add_mtd_device(mymtd);
#endif
		return 0;
	} else { /* Some boxes have 1/2 flash... */
		iounmap((void *)fsm73xx_map.map_priv_1);
		fsm73xx_map.map_priv_1
		   = (unsigned long)ioremap(WINDOW_ADDR + (WINDOW_SIZE / 2),
		                            (WINDOW_SIZE / 2));
		fsm73xx_map.size = (WINDOW_SIZE / 2);

		mymtd = do_map_probe("cfi_probe", &fsm73xx_map);
		if (mymtd) {
			mymtd->module = THIS_MODULE;

#if defined(HAVE_PARTITIONS)
			add_mtd_partitions(mymtd, fsm73xx_partitions_2,
			                   NUM_PARTITIONS);
#else
			add_mtd_device(mymtd);
#endif
			return 0;
		}
	}

	iounmap((void *)fsm73xx_map.map_priv_1);
	return -ENXIO;
}

static void __exit cleanup_fsm73xx(void)
{
	if (mymtd) {
#if defined(HAVE_PARTITIONS)
		del_mtd_device(mymtd);
#else
		del_mtd_partitions(mymtd);
#endif
		map_destroy(mymtd);
	}
	if (fsm73xx_map.map_priv_1) {
		iounmap((void *)fsm73xx_map.map_priv_1);
		fsm73xx_map.map_priv_1 = 0;
	}
}

module_init(init_fsm73xx);
module_exit(cleanup_fsm73xx);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("John W. Linville <linville@lvl7.com>");
MODULE_DESCRIPTION("MTD map driver for Netgear FSM73xx");
