/*
 * $Id: jblue.c,v 1.1 2003/12/02 21:12:07 linville Exp $
 *
 * Normal mappings of chips in physical memory
 *
 * Modified for CoreCom J-Blue partitions... -- JWL
 */

/* partition support */
#define HAVE_PARTITIONS

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/concat.h>
#include <linux/mtd/map.h>
#include <linux/config.h>
#ifdef HAVE_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define BANK1_WINDOW_ADDR	0xff000000
#define BANK1_WINDOW_SIZE	0x00800000
#define BANK0_P0_WINDOW_ADDR	0xfff00000
#define BANK0_P0_WINDOW_SIZE	0x00080000
#define BANK0_P1_WINDOW_ADDR	0xfff80000
#define BANK0_P1_WINDOW_SIZE	0x00080000

#define BUSWIDTH	1

#define APP_START	0x00000000
#define APP_LEN		0x006e0000

#define SYSTEM_START	0x006e0000
#define SYSTEM_LEN	0x00120000

#define U_BOOT_START	0x00800000
#define U_BOOT_LEN	0x00040000

#define ENV_START	0x00840000
#define ENV_LEN		0x00010000

#define VPD_START	0x00850000
#define VPD_LEN		0x00010000

#define SYSTEM_2_START	0x00860000
#define SYSTEM_2_LEN	0x000a0000

#ifdef HAVE_PARTITIONS
static struct mtd_partition jblue_partitions[] =
{
	/* app filesystem */
	{
		      name: "Application Filesystem",
		    offset: APP_START,
		      size: APP_LEN,
		mask_flags: 0
	},
	/* system (kernel + ramdisk) */
	{
		      name: "System (Kernel + Ramdisk)",
		    offset: SYSTEM_START,
		      size: SYSTEM_LEN,
		mask_flags: 0
	},
	/* u-boot */
	{
		      name: "U-Boot",
		    offset: U_BOOT_START,
		      size: U_BOOT_LEN,
		mask_flags: 0
	},
	/* u-boot environment */
	{
		      name: "U-Boot Environment",
		    offset: ENV_START,
		      size: ENV_LEN,
		mask_flags: 0
	},
	/* VPD */
	{
		      name: "FASTPATH Vital Product Data",
		    offset: VPD_START,
		      size: VPD_LEN,
		mask_flags: 0
	},
	/* system (part 2) */
	{
		      name: "System (part 2)",
		    offset: SYSTEM_2_START,
		      size: SYSTEM_2_LEN,
		mask_flags: 0
	}
};

#define NUM_PARTITIONS	(sizeof(jblue_partitions) \
                       / sizeof(struct mtd_partition))
#endif

static struct mtd_info *mymtd[3] = { NULL, };
static struct mtd_info *concat = NULL;

__u8 jblue_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 jblue_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 jblue_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void jblue_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void jblue_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void jblue_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void jblue_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void jblue_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info jblue_bank1_map = {
	name: "Bank 1",
	size: BANK1_WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: jblue_read8,
	read16: jblue_read16,
	read32: jblue_read32,
	copy_from: jblue_copy_from,
	write8: jblue_write8,
	write16: jblue_write16,
	write32: jblue_write32,
	copy_to: jblue_copy_to,
	map_priv_1: BANK1_WINDOW_ADDR
};

struct map_info jblue_bank0_p0_map = {
	name: "Bank 0 Chip 0",
	size: BANK0_P0_WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: jblue_read8,
	read16: jblue_read16,
	read32: jblue_read32,
	copy_from: jblue_copy_from,
	write8: jblue_write8,
	write16: jblue_write16,
	write32: jblue_write32,
	copy_to: jblue_copy_to,
	map_priv_1: BANK0_P0_WINDOW_ADDR
};

struct map_info jblue_bank0_p1_map = {
	name: "Bank 0 Chip 1",
	size: BANK0_P1_WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: jblue_read8,
	read16: jblue_read16,
	read32: jblue_read32,
	copy_from: jblue_copy_from,
	write8: jblue_write8,
	write16: jblue_write16,
	write32: jblue_write32,
	copy_to: jblue_copy_to,
	map_priv_1: BANK0_P1_WINDOW_ADDR
};

int __init init_jblue(void)
{
	mymtd[0] = do_map_probe("cfi_probe", &jblue_bank1_map);
	if (mymtd[0]) {
		mymtd[0]->module = THIS_MODULE;
	} else {
		return -ENXIO;
	}

	mymtd[1] = do_map_probe("jedec", &jblue_bank0_p0_map);
	if (mymtd[1]) {
		mymtd[1]->module = THIS_MODULE;
	} else {
		return -ENXIO;
	}

	mymtd[2] = do_map_probe("jedec", &jblue_bank0_p1_map);
	if (mymtd[2]) {
		mymtd[2]->module = THIS_MODULE;
	} else {
		return -ENXIO;
	}

	concat = mtd_concat_create(mymtd, 3, "J-Blue Flash");
	if (concat) {

#if defined(HAVE_PARTITIONS)
		add_mtd_partitions(concat, jblue_partitions,
		                   NUM_PARTITIONS);
#else
		add_mtd_device(concat);
#endif
		return 0;
	}

	return -ENXIO;
}

static void __exit cleanup_jblue(void)
{
	if (concat) {
#if defined(HAVE_PARTITIONS)
		del_mtd_device(concat);
#else
		del_mtd_partitions(concat);
#endif
		map_destroy(mymtd[0]);
		map_destroy(mymtd[1]);
		map_destroy(mymtd[2]);
	}
}

module_init(init_jblue);
module_exit(cleanup_jblue);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("John W. Linville <linville@lvl7.com>");
MODULE_DESCRIPTION("MTD map driver for CoreCom J-Blue");
